#define F_CPU 16000000UL
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdio.h>
#include <avr/eeprom.h> 
#include <math.h>


// serial communication library
#include "uart.h"

FILE uart_str = FDEV_SETUP_STREAM(uart_putchar, uart_getchar, _FDEV_SETUP_RW);

#define begin {
#define end }

#define Vref 5.03 

#define nopress 1
#define maybe   2
#define pressed 3
#define maybeno 4

void PIDControl(void);

unsigned char inchar;
char des_temp;
volatile unsigned char blinktime;
volatile unsigned char swtime;
char temp;
char start;
char detect_start;
char pid_flag;
volatile int Ain, AinLow ; 		//raw A to D number
volatile float  current_t_d;
volatile double  in_t,dec_t;
volatile char	in_1,in_2;
float voltage ;		//scaled input voltage


//PID setting 
float Speed,fP,fI,fD,Error,LastError,TotalError;
float ErrorArray[64];
char ErrorArrayIndex;


//lut for thermistior
//double lut_t[61] = { 3.8278,3.7815 ,   3.7343 ,   3.6862  ,  3.6371  ,  3.5873  ,  3.5369 ,   3.4856 ,   3.4336,    3.3810 ,   3.3279,
//
//    3.2743  ,  3.2200  ,  3.1656  ,  3.1106 ,  3.0555  ,   3   , 2.9444  ,  2.8888  ,  2.8330  ,  2.7771,
//
//    2.7214  ,  2.6657  ,  2.6103  ,  2.5550 ,   2.5000 ,  2.4455  ,  2.3913  ,  2.3371  ,  2.2841  ,  2.2309,
//
//    2.1789   , 2.1270  ,  2.0760  ,  2.0250  ,  1.9752  ,  1.9262  ,  1.8783  ,  1.8308  ,  1.7839  ,  1.7377,
//
//    1.6924   , 1.6480  ,  1.6048  ,  1.5620  ,  1.5197  ,  1.4789  ,  1.4387  ,  1.4003  ,  1.3619 ,   1.3244,
//
//    1.2871  ,  1.2519  ,  1.2169   , 1.1828 ,   1.1497  ,  1.1173  ,  1.0858  ,  1.0550   , 1.0251  ,  0.9960
//};



ISR (TIMER0_COMPA_vect)
begin  
	//ADC sample 
    
	swtime++; 
	if ((swtime==20) && (detect_start == 1))
	begin
	    swtime = 0;
		// wait for ADC done
		while (ADCSRA & (1<<ADSC) );
		//get the sample  
	   AinLow = (int)ADCL;
	   Ain = (int)ADCH*256; 
	   Ain = Ain + AinLow;
 		//start another conversion
 		ADCSRA |= (1<<ADSC) ; 
 		//results to hyperterm
		//program ONLY gets here after ADC ISR is done
      	voltage = (float)Ain ;
      	voltage = (voltage/1024.0)*Vref ; //(fraction of full scale)*Vref
		//best fit function
		current_t_d =  72-0.69*voltage*voltage+0.13*voltage*voltage*voltage  - 18*voltage  ; 

		in_t = floor(current_t_d);
		dec_t = current_t_d - in_t;
		in_1 = (char)in_t;
		in_2 = (char)(floor(dec_t*10));
		if(start == 1) 
            pid_flag = 1;
    end  
end // timer 0 ISR


//--PID Control algorithm-------------
void PIDControl(void)
{
	float out_NEXT;
	
	LastError=Error;	//record last error
	Error=des_temp-current_t_d;	//calculate error
	TotalError+=Error;	//for integrated part, calculate 64 error
	TotalError-=ErrorArray[ErrorArrayIndex];
	ErrorArray[ErrorArrayIndex]=Error;
	ErrorArrayIndex=(ErrorArrayIndex+1)&0b00111111;	//Index=(Index+1)%64
	out_NEXT=32+fP*Error+fD*(Error-LastError)+fI*TotalError/64;//+fD*(Error-LastError)+fI*TotalError/64;
	if (out_NEXT>=64)
		PORTC=64;
	else if (out_NEXT<=16)//threshold for negtive output
		PORTC=16;
	else
		PORTC =(char)(out_NEXT);
}

void main(void)
begin
    uart_init();
    stdout = stdin = stderr = &uart_str;
    fprintf(stdout,"Starting...\n\r");
  
    //crank up the ISRs
    //init the A to D converter 
    //channel zero/ left adj /INTERNAL Aref=Vcc
    //enable ADC and set prescaler to 1/128*16MHz=125,000
    //and clear interupt enable
    //and start a conversion
    ADCSRA = (1<<ADEN) | (1<<ADSC) + 7 ; 

    //set up timer 0
	OCR0A=249;	   			//1 mSec
    TIMSK0= (1<<OCIE0A);	//turn on timer 0 cmp match ISR
    //set prescalar to divide by 64 
    TCCR0B= 3; 	
    // turn on clear-on-match
    TCCR0A= (1<<WGM01) ;
    
    //  blink port
    DDRD = 0x04;  // and d.2 which runs another LED

	DDRC  =  0xff;//test led ports
	PORTC = 32;

	start = 0;
	detect_start = 0;
	pid_flag = 0;
        
	//-----------set pid---------
		fP = 10	    ;
		fI = 10		;
		fD = 500  	;	
	//--------------------


    sei();
	while(1)
	begin
	    temp = UCSR0A&0x80;
        if(pid_flag == 1){
			pid_flag = 0;
			PIDControl();
		} 
		 
		if (temp)
		begin
			inchar = UDR0;
			if (inchar=='s')
			begin  
				detect_start = 1;
			    fprintf(stdout,"%d %d  \r", in_1,in_2) ;
			end else if (inchar >=48 && inchar < 114)//second higher bits is check bit
			begin
				//fprintf(stdout,"%c \r", 'k');
				start = 1;
			    des_temp = inchar-48;
			end else if (inchar == 't')
			begin
				start = 0;
				detect_start = 0;
				PORTC = 32;
			end
		end		
	end  //while(1)
end //main
